############################################################################
# Joshua R. Boverhof, LBNL
# Monte M. Goode, LBNL
# David W. Robertson, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, re, weakref, string, warnings
import ZSI
from ZSI import _seqtypes
from ZSI.TC import _is_xsd_or_soap_ns, TypeCode
from ZSI.TCcompound import Struct
from ZSI.wstools import WSDLTools
from ZSI.wstools.Utility import SplitQName
from ZSI.wstools.Namespaces import SOAP, SCHEMA
from ZSI.wstools.XMLSchema import LocalComplexType
from ZSI.typeinterpreter import BaseTypeInterpreter
from ZSI.wsdlInterface import ZSIWsdlAdapter, ZSISchemaAdapter

# $Id: wsdl2python.py,v 1.62.2.16 2004/07/20 06:10:12 boverhof Exp $

"""
wsdl2python:
    This module generates an client interface module and a module containing
    all of the types (typecodes) for use with the zsi library.  

Arch:
    WriteServiceModule
      -Takes a wsdl object and creates the client interface and typecodes for 
       the service.
      -Handles writing of import complexities and class definition order.
      -Delegates to ServiceDescription interpretation of service definition.
      -Delegates to SchemaDescription interpretation of schema definition.

    ServiceDescription
      -Interprets the service definition, creates client interface and
       port descriptions code.
      -Delegates to MessageWriter

       MessageWriter
         -Generates a message's class definition.

          PartWriter
            -Generates a message part's description.


    SchemaDescription
      -Interprets the schema definition, creates typecode module code.

       TypeWriter
         -Generates a type's description/typecode.
"""


ID1 = '    '
ID2 = 2*ID1
ID3 = 3*ID1
ID4 = 4*ID1
ID5 = 5*ID1
ID6 = 6*ID1

IDSP2 = '  '

DEBUG  = 0
NSDICT = {}

def print_debug(msg, level=1, *l, **kw):
    if DEBUG >= level:
        for i in l:
            msg += '\n*\t' + str(i)
        for k,v in kw.items():
            msg += '\n**\t%s: %s' %(k,v)
        print "WSDL2PYTHON(%d): %s" %(level,msg)

def _getLiteralFromTypeClassVar(classvar):
    '''This function is for taking a class definition that has
    incorrectly been defined with a "type" class variable.  It 
    basically changes the class variable "type" to "literal".
    WARNING: Prone to errors.
    '''
    lines = classvar.split('\n')
    for indx in range(len(lines)):
        type_line = lines[indx]
        if type_line.find('type = (schema') > 0: break
    else:
        raise WsdlGeneratorError, 'Cannot find type class variable'
    literal_line = re.sub('type = \(schema,', 'literal = ', type_line)
    lines[indx] = literal_line[:literal_line.rfind(")")]
    value = ''
    for line in lines:
        #value = '\n' + lines[0] + '\n' + lines[1] + '\n' + lines[2] + '\n' + lines[3] + '\n'
        value = '%s\n%s' %(value,line)
    value += '\n'
    return value

def nonColonizedName_to_moduleName(name):
    """Take a non-colonized name, create a legal module name.
       Replace . -- _
    """
    return re.sub('\.', '_', name)

def textProtect(s):
    """process any strings we cant have illegal chracters in"""
    return re.sub('[-./:?*# ]', '_', s)


class WsdlGeneratorError(Exception):
    pass

class Wsdl2PythonError(Exception):
    pass

class MessageContainer:
    '''
    '''
    class IntHolder(int): pass
    class StrHolder(str): pass
    class UnicodeHolder(unicode): pass
    class LongHolder(long): pass
    class FloatHolder(float): pass
    class BoolHolder(int): pass
    #class TupleHolder(tuple): pass
    #class ListHolder(list): pass
    typecode = None
    pyclass_list = [IntHolder,StrHolder,UnicodeHolder,LongHolder,
        FloatHolder,BoolHolder,]

    def __setattr__(self, key, value):
        '''wrap all primitives with Holders if present.
        '''
        if type(value) in _seqtypes:
            value = list(value)
            for indx in range(len(value)):
                try:
                    item = self._wrap(value[indx])
                except TypeError, ex:
                    pass
                else:
                    value[indx] = item
        elif type(value) is bool:
            value = BoolHolder(value)
        else:
            try:
                value = self._wrap(value)
            except TypeError, ex:
                pass
        self.__dict__[key] = value

    def _wrap(self, value):
        '''wrap primitive, return None
        '''
        if value is None: 
            return value
        for pyclass in self.pyclass_list:
            if issubclass(value.__class__, pyclass): break
        else:
            raise TypeError, 'MessageContainer does not know about type %s' %(type(value))
        return pyclass(value)

    def setUp(self, typecode=None, pyclass=None):
        '''set up all attribute names (aname) in this python instance.
        If what is a Struct or a simpleType w/attributes instantiate 
        a new MessageContainer, else set attribute aname to None.
        '''
        if typecode is None: typecode = self.typecode
        if not isinstance(typecode, TypeCode):
            raise TypeError, 'typecode must be a TypeCode class instance'

        if isinstance(typecode, ZSI.TCcompound.Struct):
            if typecode.has_attributes() is True:
                setattr(self, typecode.attrs_aname, {})
            if typecode.mixed is True:
                setattr(self, typecode.mixed_aname, None)
            for what in typecode.ofwhat:
                setattr(self, what.aname, None)
                if isinstance(what, ZSI.TCcompound.Struct):
                    setattr(self, what.aname, MessageContainer())
                    getattr(self, what.aname).setUp(typecode=what)
        else:
            raise TypeError, 'Primitive type'


class SchemaBase:
    '''Helper class for generated code
    '''
    schema = None

    def get_aname(name):
        '''Class method for constructing an aname from element
        declaration class information.
        '''        
        if name is None:
            return None
        return '_%s' %name
    get_aname = staticmethod(get_aname)

    def get_nsuri(cls, **kw):
        namespaceURI = getattr(cls, 'schema')
        return namespaceURI
    get_nsuri = classmethod(get_nsuri)

    def get_variable(self, pyclass):
        '''Find instance of pyclass in ofwhat, and
        return the matching instance variable.
        '''
        if isinstance(self, Struct):
            for what in self.ofwhat:
                if isinstance(what, pyclass): break
            else:
                raise EvaluateException, ''
        else:
            raise EvaluateException, ''
        aname = getattr(what, 'aname', None)
        if aname is None:
            raise EvaluateException, ''
        return getattr(self,aname)


class ElementDeclaration(SchemaBase):
    '''Helper class for generated code
    '''
    literal = None

    def get_pname(cls, **kw):
        '''Class method for retrieving element namespaceURI/name
        information, used to construct a pname. 
            Key Word:
                encoded -- encoded namespace (ignore for now).
        '''        
        namespaceURI = getattr(cls, 'schema')
        name = getattr(cls, 'literal')
        return (namespaceURI,name)
    get_pname = classmethod(get_pname)


class TypeDefinition(SchemaBase):
    '''Helper class for generated code
    '''
    pass


class NamespaceHash:
    """Keep track of the various namespaces used in a service and assign
    aliases to said namespaces for brevity.  A lookup table
    """
    NSCOUNT = 1
    NSDICT = {}
    NSORDER = []

    def __init__(self):
        pass

    def add(self, ns):

        if not NamespaceHash.NSDICT.has_key(ns):
            NamespaceHash.NSDICT[ns] = ( self.namespace_to_moduleName(ns),
                                         'ns%s' % NamespaceHash.NSCOUNT )
            NamespaceHash.NSORDER.append(ns)
            NamespaceHash.NSCOUNT += 1


    def getModuleName(self, ns):

        if NamespaceHash.NSDICT.has_key(ns):
            return NamespaceHash.NSDICT[ns][0]
        else:
            raise WsdlGeneratorError, 'could not retrieve mod name for %s' % ns

    def getAlias(self, ns):

        if NamespaceHash.NSDICT.has_key(ns):
            return NamespaceHash.NSDICT[ns][1]
        else:
            self.dump()
            raise WsdlGeneratorError, 'could not retrieve alias for %s' % ns

    def getNSList(self):
        return NamespaceHash.NSORDER


    def namespace_to_moduleName(self, n):
        name = n
        
        if name.startswith('http://'):
            name = name[7:]

        name = textProtect(name)

        if name.endswith('_'):
            name = name[:-1]
        
        return name

    def dump(self):
        print NamespaceHash.NSDICT

    def reset(self):
        NamespaceHash.NSCOUNT = 1
        NamespaceHash.NSDICT = {}
        NamespaceHash.NSORDER = []


class StringWriter:
    def __init__(self, val=None):
        self.data = []
        if val:
            self.data.append(val)
            
    def set(self, val):
        if self.data:
            # in some cases the empty list reassignment fails, so....
            self.data = None
            self.data = []
            
        self.data.append(val)
        
    def write(self, val):
        self.data.append(val)
        
    def __iadd__(self, val):
        self.data.append(val)
        return self

    def getvalue(self):
        if self.data:
            return ''.join(self.data)
        else:
            return ''

        
class WriteServiceModule:
    """Takes a wsdl object and creates the client interface and typecodes for 
       the service.
    """
    def __init__(self, wsdl, importlib=None, typewriter=None, addressing=False, notification=False, configParser=None):
        """wsdl - wsdl.Wsdl instance
        """
        self.wsdl  = wsdl
        self._wa = ZSIWsdlAdapter( self.wsdl )
        self.nsh = NamespaceHash()
        self.tns_wrote = {}
        self.tns_imported = {}
        self._nscount = 1
        self._nsdict  = {}
        self._importlib  = importlib
        self._typewriter = typewriter
        self._addressing = addressing
        self._configParser = configParser
        self._notification = notification
        self.typeDict = {}

    def write(self, schemaOnly=False):
        """Write schema instance contents w/respect to dependency requirements, 
           and create client interface.  Not guaranteed to work for mutual 
           depenedencies.
        """

        f_types, f_services = self.get_module_names()
        hasSchema = len(self._wa.getSchemaDict())
        if hasSchema:
            fd = open(f_types + ".py", 'w+')
            self.write_service_types(f_types, fd)
            fd.close()

        if schemaOnly:
            return

        fd = open(f_services + ".py", 'w+')
        self.write_services(f_types, f_services, fd, hasSchema)
        fd.close()

        
    def get_module_names(self):
        if self._wa.getName():
            name = nonColonizedName_to_moduleName( self._wa.getName() )
        else:
            raise WsdlGeneratorError, 'could not determine a service name'

        name = SplitQName(name)[1]
        f_types = '%s_services_types' % name
        f_services = '%s_services' % name
        return f_types, f_services


    def write_service_types(self, f_types, fd):
        header = '%s \n# %s.py \n# generated by %s \n# \n# \n%s\n\n'\
	          %('#'*50, f_types, self.__module__, '#'*50)

        imports = ['\nimport ZSI', 
            '\nimport ZSI.TCcompound',
            '\nfrom ZSI.wsdl2python import ElementDeclaration,TypeDefinition',
            '\nfrom ZSI.TC import _add_namespace_container',
        ]
        #imports.append('\nfrom ZSI.TCcompound import Struct')

        if self._importlib:
            exec( 'import %s' % self._importlib )
            exec( 'obj = %s' % self._importlib )
            if hasattr( obj, 'typeimports' ):
                imports += obj.typeimports

	fd.write(header)
	fd.writelines(imports)
	fd.write('\n'*2)

        for schema in self._wa.getSchemaDict().values():
            self.write_dependent_schema(schema, fd)
            

    def write_services(self, f_types, f_services, fd, hasSchema):
        header = '%s \n# %s.py \n# generated by %s \n# \n# \n%s\n\n'\
	          %('#'*50, f_services, self.__module__, '#'*50)

	fd.write(header)
        
        for service in self._wa.getServicesList():
            #sd = ServiceDescription(self._addressing, self._notification, configParser=self._configParser)
            sd = ServiceDescription(self._addressing)
            if hasSchema:
                sd.imports += ['\nfrom %s import *' % f_types]
                for ns in self.nsh.getNSList():
                    sd.imports += ['\nfrom %s import \\\n    %s as %s' \
                                  % ( f_types,
                                      self.nsh.getModuleName(ns),
                                      self.nsh.getAlias(ns)) ]

            sd.fromWsdl(service, self.typeDict)
            
            if self._importlib:
                exec( 'import %s' % self._importlib )
                exec( 'obj = %s' % self._importlib )
                if hasattr(obj, 'clientimports' ):
                    sd.imports += obj.clientimports
                    

            sd.write(fd)

        fd.write('\n')


    def write_dependent_schema(self, schema, fd):
        """Write schema instance contents w/respect to dependency requirements.
           First write any schema that is imported directly into current schema,
           then check current schema's xmlns and see if any of these represent
           currently held schema instances.

           schema -- schema.Schema instance
           fd -- file descriptor
        """

        # check imports
        for ns in schema.getImports():
            if self.tns_imported.has_key(ns):
                #raise WsdlGeneratorError,\
                #      'suspect circular import of %s - not suported' % ns
                pass
            if ns not in SCHEMA.XSD_LIST and \
               ns not in [SOAP.ENC]:
                self.tns_imported[ns] = 1
            if self._wa.getSchemaDict().has_key(ns) and \
                   (not self.tns_wrote.has_key(ns)):
                self.write_dependent_schema(self._wa.\
                                            getSchemaDict().get(ns), fd)
            if schema.getImportSchema(ns) and \
               (not self.tns_wrote.has_key(ns)):
                self.write_dependent_schema(schema.getImportSchema(ns), fd)

        # next...xmlns
        for ns in schema.getXmlnsDict().values():
            if self._wa.getSchemaDict().has_key(ns) and \
                   (not self.tns_wrote.has_key(ns)) and \
                   (ns != schema.getTargetNamespace()) and \
                   (not schema.getTargetNamespace() in \
                    self._wa.getSchemaDict().get(ns).getXmlnsDict().values()):
                self.write_dependent_schema(self._wa.getSchemaDict().get(ns),
                                            fd)
        
        # fall through
        else:
            if not self.tns_wrote.has_key(schema.getTargetNamespace()):
                self.nsh.add(schema.getTargetNamespace())
                self.tns_wrote[schema.getTargetNamespace()] = 1
                if self._importlib and self._typewriter:
                    alternateWriter = ( self._importlib, self._typewriter )
                else:
                    alternateWriter = None
                sd = SchemaDescription()
                sd.fromWsdl(schema, alternateWriter)
                sd.write(fd)
                    # typeDict used when generating 'services' file
                    # to generate docstrings for each operation
                self.typeDict.update(sd.typeDict)

    def __del__(self):
        self.nsh.reset()
 
class ServiceDescription:
    """Generates client interface.  Writes out an abstract interface, 
    a locator class, and port classes.

    class variables:
       wsAddressing -- Turn on WS-Addressing

    """
    def __init__(self, addressing=False):
        '''Instance variables:
             wsAddressing -- Turn on WS-Addressing
             wsNotification -- Turn on WS-Notification
             configParser -- ConfigParser instance specifying special operations,
                [section] operations
                  create
                  destroy
                  Subscribe
        '''
	self.imports = []
        self.messages = None
        self.serviceBindings = None
	self.serviceInterface = None
	self.serviceLocator = None
        self.wsAddressing = addressing

    def write(self, fd=sys.stdout):
        """Write service instance contents.  Must call fromWsdl with
           service instance before calling write.
        """
        if self.messages is None:
            raise WsdlGeneratorError, 'must call fromWsdl before write'
	fd.writelines(self.imports)
	fd.write(self.serviceInterface)
	fd.write(self.serviceLocator)
        fd.write(self.serviceBindings)
        keys = self.messages.keys()
        keys.sort()
        [fd.write(str(self.messages[k])) for k in keys]
            
    def fromWsdl(self, service, typeDict=None):
        """service -- wsdl.Service instance
        """
	self.imports += '\nimport urlparse, types'
	self.imports += '\nfrom ZSI.TCcompound import Struct'
	self.imports += '\nfrom ZSI import client'
	self.imports += '\nimport ZSI'
        self.messages = {}
	self.serviceInterface = '\n\nclass %sInterface:\n' % \
                                (service.getName())
	self.serviceLocator = '\n\nclass %sLocator(%sInterface):\n' % \
                              (service.getName(), service.getName())
        self.serviceBindings = ''

        hasSoapBinding = False

        defs = 'defs'
        for p in service.getPortList():
            port = p._port
            binding = port.getBinding()
            portType = binding.getPortType()
            rProp = portType.getResourceProperties()

            # -----------------------------
            # REQUIRED: WSDLv1.2 Bindings
            #  skip port if not present
	    #  port.soap
	    #    soap:address  location
            # -----------------------------
        
            hasSoapAddress = False
            soapAddress    = None

            for e in p.getExtensions():
                if isinstance(e,
                              ZSI.wsdlInterface.ZSISoapAddressAdapter):
                    hasSoapAddress = True
                    soapAddress = e

            if not hasSoapAddress:
                continue
            else:
                hasSoapBinding = True
                
            # -----------------------------
	    #  class variable address

	    self.serviceLocator += ID1 
            self.serviceLocator += '%s_address = "%s"\n' % \
                                   ( p.getBinding().getPortType().getName(),
                                     soapAddress.getLocation() )

            # -----------------------------
	    #  getPortTypeAddress

	    self.serviceLocator += ID1 
	    self.serviceLocator +=  'def get%sAddress(self):\n' % \
                                   ( p.getBinding().getPortType().getName() )
	    self.serviceLocator += ID2 
	    self.serviceLocator +=  'return %sLocator.%s_address\n\n' % \
                                   ( service.getName(),
                                     p.getBinding().getPortType().getName() )


            # -----------------------------
            #  getPortType(portAddress)

	    self.serviceInterface += \
                                  ID1 + \
                                  'def get%s(self, portAddress=None, **kw):\n'\
                                  % ( p.getBinding().getPortType().getName() )
	    self.serviceInterface += ID2 
	    self.serviceInterface += 'raise NonImplementationError, "method not implemented"\n'

	    self.serviceLocator += \
                                ID1 + \
                                'def get%s(self, portAddress=None, **kw):\n' \
                                % ( p.getBinding().getPortType().getName() )
	    self.serviceLocator += ID2 
	    self.serviceLocator += 'return %sSOAP(portAddress or %sLocator.%s_address, **kw)\n' \
                                   % ( p.getBinding().getName(),
                                       service.getName(),
                                       p.getBinding().getPortType().getName() )

	    myBinding = {'class':'\n\nclass %sSOAP:' % \
                         (p.getBinding().getName()), 'defs':{} }
            if rProp is not None:
	        myBinding['class'] += '\n%s#ResourceProperties = ("%s","%s")'%(ID1,rProp[0],rProp[1])



            # -----------------------------
            # Methods
            # -----------------------------
            defs = 'defs'
            method = '__init__'
            if self.wsAddressing is True:
                myBinding[defs][method] =  '\n%sdef __init__(self, addr, endPointReference=None, **kw):'%ID1
            else:
                myBinding[defs][method] =  '\n%sdef __init__(self, addr, **kw):'%ID1
            myBinding[defs][method] += '\n%snetloc = (urlparse.urlparse(addr)[1]).split(":") + [80,]'%ID2
            myBinding[defs][method] += '\n%sif not kw.has_key("host"):'%ID2
            myBinding[defs][method] += '\n%skw["host"] = netloc[0]'%ID3
            myBinding[defs][method] += '\n%sif not kw.has_key("port"):'%ID2
            myBinding[defs][method] += '\n%skw["port"] = int(netloc[1])'%ID3
            myBinding[defs][method] += '\n%sif not kw.has_key("url"):'%ID2
            myBinding[defs][method] += '\n%skw["url"] =  urlparse.urlparse(addr)[2]'%ID3
            if rProp:
	        myBinding[defs][method] += '\n%skw["ResourceProperties"] = ("%s","%s")'%(ID2,rProp[0],rProp[1])

            if self.wsAddressing is True:
                myBinding[defs][method] += '\n%sself.binding = client.Binding(wsAddressing=True, **kw)'%(ID2)
                myBinding[defs][method] += '\n%sself.endPointReference = endPointReference'%(ID2)
            else:
                myBinding[defs][method] += '\n%sself.binding = client.Binding(wsAddressing=False, **kw)'%(ID2)

            for op in p.getBinding().getPortType().getOperationList():
                bopName = op.getName()
                for bop in binding.operations:
                    if bop.name == bopName: break
                else:
                    raise RuntimeError, 'No bop name(%s)' %bopName
                
                # -----------------------------
                #  REQUIREMENTS SOAP Bindings
	        #  port.binding.operations.soap[binding,operations]
	        #  port.binding.soap[binding]
                #        style, transport
                # -----------------------------
                style      = None
                transport  = None
                soapAction = None
                use           = None
                namespace     = None
                encodingStyle = None

                #Soap Binding
                sob = binding.findBinding(WSDLTools.SoapBinding)
                if sob:
                    transport,style = sob.transport,sob.style

                #Soap Operation Bindings
                soapOpBinding = bop.findBinding(WSDLTools.SoapOperationBinding)
                if soapOpBinding:
                    soapAction,style = soapOpBinding.soapAction,soapOpBinding.style
                elif bop.findBinding(WSDLTools.HttpOperationBinding) is not None:
                       raise RuntimeError, 'not supporting HTTP bindings'

                #SoapBodyBinding
                if bop.input is not None:
                    soapBodyBinding = bop.input.findBinding(WSDLTools.SoapBodyBinding)
                    #self.parts = parts
                    encodingStyle = soapBodyBinding.encodingStyle 
                    namespace = soapBodyBinding.namespace 
                    use = soapBodyBinding.use

                operation = p.getBinding().getOperationDict().get(op.getName())
                if operation.getSoapBinding():
                    #style      = operation.getSoapBinding().getStyle()
                    #transport  = operation.getSoapBinding().getTransport()
                    soapAction = operation.getSoapBinding().getAction()
                else:
                    #style      = p.getBinding().getSoapBinding().getStyle()
                    #transport  = p.getBinding().getSoapBinding().getTransport()
                    pass

                if operation.getInput().getSoapBody():
                    use           = operation.getInput().getSoapBody().getUse()
                    namespace     = operation.getInput().getSoapBody().getNamespace()
                    encodingStyle = operation.getInput().getSoapBody().getEncoding()


                # -----------------------------
		# Methods
                # -----------------------------
                if op.getInput():
                    inputName = op.getInput().getMessage().getName()
                    self.messages[inputName] = self.__class__.MessageWriter()
                    self.messages[inputName].fromMessage(op.getInput().getMessage(), 
                                          op.getName(), namespace, style, use)
                
                    outputName = None
                    if op.getOutput() and op.getOutput().getMessage():
                        outputName = op.getOutput().getMessage().getName()
                        self.messages[outputName] = \
                                      self.__class__.MessageWriter()
                        self.messages[outputName].\
                                      fromMessage(op.getOutput().getMessage(), 
                                                  outputName,
                                                  namespace,
                                                  style,
                                                  use, op)

		    myBinding[defs][op.getName()] = \
                                                    '%sdef %s(self, request):' % \
                                                    (ID1, textProtect(op.getName()))
                    myBinding[defs][op.getName()] += '\n%s"""\n' % ID2
                    myBinding[defs][op.getName()] += self.messages[inputName].\
                                         docString(typeDict, inputName, True)
                    myBinding[defs][op.getName()] += '\n'
                    if outputName:
                        myBinding[defs][op.getName()] += self.messages[outputName].\
                                         docString(typeDict, outputName, False)
                    myBinding[defs][op.getName()] += '%s"""\n' % ID2
                    
                    # checking to handle the special case of an
                    # element declaration of a primitive type.
                    # you will find these in document/literal ops.
                    kwstring = None
                    simpleType = self.isSimpleElementDeclaration(op, input=True)

                    if simpleType:
                        kwstring = "\n%skw = {'requestclass': %s}"%(ID2, inputName )
                        myBinding[defs][op.getName()] += '\n%sif not type(request) == %s:' %( ID2, simpleType)
                    else:
                        kwstring = '\n%skw = {}' % ID2
                        myBinding[defs][op.getName()] +=\
                                                        '\n%sif not isinstance(request, %s) and\\\n%snot issubclass(%s, request.__class__):'\
                                                        %(ID2, inputName,
                                                          ID3, inputName)
		    myBinding[defs][op.getName()] +=\
                                                    '\n%sraise TypeError, %s' % \
                                                    (ID3, r'"%s incorrect request type" %(request.__class__)')

                    myBinding[defs][op.getName()] += kwstring
                    if self.wsAddressing is True:
                        wsaction = portType.operations[op.getName()].getInputAction()
                        myBinding[defs][op.getName()] += '\n%swsaction = "%s"' %(ID2, wsaction)
			myBinding[defs][op.getName()] += \
                            '\n%sself.binding.Send(None, None, request, soapaction="%s", wsaction=wsaction, endPointReference=self.endPointReference, **kw)'\
                            %(ID2, soapAction)
                    else:
                        myBinding[defs][op.getName()] +=\
                            '\n%sself.binding.Send(None, None, request, soapaction="%s", **kw)' %(ID2, soapAction)

                    if outputName:
                        if self.wsAddressing is True:
                            wsaction = portType.operations[op.getName()].getOutputAction()
                            myBinding[defs][op.getName()] += '\n%swsaction = "%s"' %(ID2, wsaction)
			    myBinding[defs][op.getName()] += '\n%sresponse = self.binding.Receive(%s(), wsaction=wsaction)' %(ID2, outputName)
                        else:
			    myBinding[defs][op.getName()] += '\n%sresponse = self.binding.Receive(%s())' %(ID2, outputName)

                        simpleType = self.isSimpleElementDeclaration(op, input=False)
                        if simpleType:
                            myBinding[defs][op.getName()] += '\n%sif not type(response) == %s:' %( ID2, simpleType)
                        else:
                            myBinding[defs][op.getName()] +=\
                                '\n%sif not isinstance(response, %s) and\\\n%snot issubclass(%s, response.__class__):'\
                                %(ID2, outputName, ID3, outputName)
		        myBinding[defs][op.getName()] +=\
                            '\n%sraise TypeError, %s' % \
                            (ID3, r'"%s incorrect request type" %(response.__class__)')
		        myBinding[defs][op.getName()] += '\n%sreturn response' %ID2
                        
		elif op.output:
		    pass
		else:
		    raise WsdlGeneratorError,\
                          'Operation w/o input and/or output'

            else:
	        self.serviceBindings += myBinding['class']
		self.serviceBindings += "\n%s\n" \
                                        % (myBinding[defs]['__init__'])
		del myBinding[defs]['__init__']

                keys = []
                #sort out all _methods
                for k in myBinding[defs].keys():
                    if not k.startswith('_'): 
                        keys.append(k)
                    else: 
                        d = myBinding[defs][k]
		        self.serviceBindings += "\n%s\n" %(d)

                keys.sort()
                #sort the rest alphabetically
                for mn in keys:
                    d = myBinding[defs][mn]
		    self.serviceBindings += "\n%s\n" %(d)
		else:
	            self.serviceBindings += "\n"

        if not hasSoapBinding:
           raise WsdlGeneratorError, 'no soap bindings available for service %s' % service.getName()
        return

    def isSimpleElementDeclaration(self, op, input=True):
        '''JRB: This method is really ugly and very error prone.  
        It returns True/False/StringType/IntType
        '''
        if input is True:
            message = op.getInput().getMessage()
        else:
            message = op.getOutput().getMessage()
        if len( message.getPartList() ) == 1:
            prt = message.getPartList()[0]

            if prt.getElement():
                #isBasicElement is returning StringType/Int, etc.  
                return prt.getElement().isBasicElement()

        return False

    class MessageWriter:
        """Generates a class representing a message.
        """
	def __init__(self):
            self.nsh = NamespaceHash()
	    self.typecode = None
            self.typeList = []

	def __str__(self):
	    return self.typecode

	def fromMessage(self, message, name, namespace, style, use, op=None):
	    """message -- wsdl.Message instance
               name -- operation name
               namespace -- soap binding namespace
               style -- 'rpc' or 'encoded'
               use -- 'document' or 'literal'
            """
            if use == 'encoded':
                #self.fromMessageEncoded(message, name, namespace, style, op)
                raise RuntimeError, "NOT SUPPORTED"
            elif use == 'literal':
                self.fromMessageLiteral(message, name, namespace, style)
            else:
                raise WsdlGeneratorError, 'unsupported use=%s' %(use)

        '''
        def fromMessageEncoded(self, message, name, namespace, style, op):
	    l = []

            classname = ''

            if message.isOutput() and style == 'rpc':
                if not op:
                    # this is mostly a temp bulletproofing exception.
                    raise WsdlGeneratorError, 'operation object is none - bad'
                else:
                    name = op.getName() + 'Response'

	    self.typecode = '\n\nclass %s (ZSI.TCcompound.Struct): '\
                            %(message.getName())
            self.typecode += '\n%sdef __init__(self, name=None, ns=None):'\
                             %(ID1)

	    for p in message.getPartList():
                if p.getType():
                    tns = p.getType().getTargetNamespace()
                    tp = self.__class__.PartWriter()
                    tp.fromPart(p)
                    if tp.typecode[0][0:3] == 'ZSI':
                        l += tp.typecode
                        self.typecode += '\n%sself._%s = None'\
                                         % ( ID2, p.getName() )
                            # for later use in docstring generation
                        self.typeList.append(['_' + tp.docCode[0], tp.docCode[1], False])
                    else:
                        qualifiedtc = tp.typecode[0:]
                        idx = qualifiedtc[0].find('(')
                        qualifiedtc[0] = self.nsh.getAlias(tns) + \
                                         '.' + qualifiedtc[0][0:idx] + \
                                         '_Def' + qualifiedtc[0][idx:]
                        l += qualifiedtc
                        defclass = tp.typecode[0][0:]
                        defclass = defclass[0:defclass.find('(')] + '_Def()'
                        self.typecode += '\n%sself._%s = %s.%s'\
                                         % ( ID2, p.getName(),
                                             self.nsh.getAlias(tns),
                                             defclass )
                        self.typeList.append(['_' + tp.docCode[0],
                                                   '%s.%s_Def' % (self.nsh.getAlias(tns), tp.docCode[1]), False])
                elif p.getElement():
                    raise WsdlGeneratorError, 'Bad encoding'
                else:
                    raise WsdlGeneratorError, 'FIX'
	    else:
                self.typecode += '\n\n%soname = None' % ID2
                self.typecode += '\n%sif name:' % ID2
                self.typecode += '\n%soname = name' % ID3
                self.typecode += '\n%sif ns:' % ID3
                self.typecode += "\n%soname += ' xmlns=\"%%s\"' %% ns" % ID4
		tcs = ''
		for i in l: tcs += (i + ',')
                self.typecode += '\n%sZSI.TC.Struct.__init__(self, %s, [%s], pname=name, aname="_%%s" %% name, oname=oname )'\
                                 %( ID3, message.getName(), tcs )

            # do the wrapper do go with the message

            if style == 'document':
                name = None
            elif style == 'rpc':
                name = "'" + name + "'"
            else:
                raise WsdlGeneratorError, 'incorrect document type -> ?'

            self.typecode += '\n\n'
            self.typecode += 'class %sWrapper(%s):\n' \
                             % (message.getName(), message.getName())
            self.typecode += '%s"""wrapper for %s:encoded message"""\n\n' \
                             % ( ID1, style )

            self.typecode +=\
                          "%stypecode = %s(name=%s, ns='%s')"\
                          % ( ID1, message.getName(), name, namespace )
            self.typecode += '\n%sdef __init__( self, name=None, ns=None, **kw ):'\
                             % ( ID1 )
            self.typecode += "\n%s%s.__init__( self, name=%s, ns='%s' )"\
                             % (ID2, message.getName(), name, namespace)
        '''
            

        def fromMessageLiteral(self, message, name, namespace, style):

	    l = []

	    #self.typecode = '\n\nclass %s: '%message.getName()
	    #self.typecode += '\n%sdef __init__(self, **kw):'%ID1

	    parts = message.getPartList()
	    if len(parts) > 1:
                raise WsdlGeneratorError, \
                    'Not Supporting doc/literal messages with multiple parts'
	    elif len(parts) == 0:
                raise WsdlGeneratorError, \
                    'Must specify a part for a doc/literal message'

	    for p in parts:
		if p.getElement():
		    tp = self.__class__.PartWriter()
		    tp.fromPart(p)
		    if tp.name:
                        nsp = self.nsh.getAlias(p.getElement().getTargetNamespace())
                        self.typecode = '\n\nclass %s(%s.%s_Dec): ' %(message.getName(),nsp,tp.name)
                        #self.typecode += '\n%sif not hasattr( %s.%s(), "typecode" ):' % (ID1, nsp, tp.name + '_Dec')
                        self.typecode += '\n%stypecode = %s.%s_Dec()' %(ID1,nsp,tp.name)
			self.typecode += '\n\n%sdef __init__(self, name=None, ns=None):'%ID1
                        self.typecode += '\n%s%s.%s_Dec.__init__(self, name=None, ns=None)'%(ID2,nsp,tp.name)
                        # for later use in docstring generation
                        self.typeList.append(['literal', '%s.%sLOCAL_Def' % (nsp, tp.name), False])
		    else:
			self.typecode = '\n\nclass %s: ' % (message.getName())
			self.typecode += '\n%sdef __init__(self, name=None, ns=None): pass' %(ID1)
		    break
		else:
                    raise WsdlGeneratorError, \
                        'Not Supporting doc/literal messages with <part type>'

            # do the wrapper to go w/the message
            #if style == 'document':
            #    name = None
            #elif style == 'rpc':
            #    name = "'" + name + "'"
            #else:
            #    raise WsdlGeneratorError, 'incorrect document type -> ?'

            #self.typecode += '\n\n'
            #self.typecode += 'class %sWrapper(%s):\n' \
            #                 % (message.getName(), message.getName())
            #self.typecode += '%s"""wrapper for %s:literal message"""\n\n' \
            #                 % (ID1, style )

            #self.typecode += '%stypecode = %s( name=%s, ns=None ).typecode'\
            #                 % (ID1, message.getName(), name )
            #self.typecode += '\n%sdef __init__( self, name=None, ns=None, **kw ):'\
            #                 % ID1
            #self.typecode += '\n%s%s.__init__( self, name=%s, ns=None )' \
            #                 % (ID2, message.getName(), name )


        def docString(self, allTypeDict, messageName, isInput):
            """Generates a docstring giving the input and output
               message names, and their associated parameters and
               parameter sub-fields.
            """
            docList = []
            if self.typeList:
                self.typeList.sort()

            for item in self.typeList:
                if item[1]:
                    item[1] = item[1].replace('LOCAL_Def', '_Def')
                else:
                    item[1] = 'Any'
            alreadyListed = []
            paramList = self.recurseTypeList(allTypeDict, self.typeList,
                                                       1, alreadyListed)
            if paramList:
                if paramList[0].startswith('*Simple*'):
                    startIndex = len('*Simple*') + 1
                    if isInput:
                        docList.append('%s@param: request is %s\n' % (ID2, paramList[0][startIndex:]))
                    else:
                        docList.append('%s@return: response is %s\n' % (ID2, paramList[0][startIndex:]))
                    return ''.join(docList)

            if isInput:
                docList.append('%s@param: request to %s' % (ID2, messageName))
            else:
                docList.append('%s@return: response from %s' % (ID2, messageName))
            if self.typeList:
                docList.append('::\n')
            docList.extend(paramList)
            return ''.join(docList)


        def recurseTypeList(self, allTypeDict, typeList, level, alreadyListed):
            """Recurse through a dictionary of type lists until reach
               Python types.
            """
            if level > 12:  # just in case
                return
            strList = []
            for item in typeList:
                if item[0].startswith('_'):
                    if not item[0].startswith('__param'):
                        indent = ID2 + IDSP2*level
                        strList.append('%s%s: %s' % (indent, item[0], item[1]))
                        if item[2]:
                            strList.append(', minOccurs')
                        strList.append('\n')
                    else:
                        return ['*Simple* %s' % item[1]]
                if allTypeDict.has_key(item[1]):
                    if item[1] not in alreadyListed:
                        alreadyListed.append(item[1])
                        strList.extend(self.recurseTypeList(allTypeDict,
                               allTypeDict[item[1]], level+1, alreadyListed))
            return strList


	class PartWriter:
            """Generates a string representation of a typecode representing
               a <message><part>
            """
	    def __init__(self):
		self.typecode = None
                self.tns = None
		self.name = None
                self.docCode = []

	    def __recurse_tdc(self, tp):
                """tp -- schema.TypeDescriptionComponent instance
                """
		tp.type
		if isinstance(tp.type, TypeDescriptionComponent):
		    tp.type
		    tp = self.__recurse_tdc(tp.type)
		else:
		    return tp.type

	    def fromPart(self, part):
                """part -- wsdl.Part instance
                """

		bti = BaseTypeInterpreter()

		if part.getType():
		    tp = part.getType()
		elif part.getElement():
		    tp = part.getElement()
		    self.name = tp.getName()
		    return
                else:
                    raise WsdlGeneratorError, 'whoa!  part typing problem!'

		self.typecode = []

                self.tns = tp.getTargetNamespace()

                if not isinstance(tp, ZSI.wsdlInterface.ZSITypeAdapter):
                    raise TypeError, 'not a type adapter'

		elif tp.isSimpleType():
		    if tp.getQName():
			tpc = bti.get_typeclass(tp.getQName(),
                                                tp.getTargetNamespace())
                        self.docCode.append('%s' % part.getName())
                        self.docCode.append(bti.get_pythontype(tp.getQName(),
                                                tp.getTargetNamespace()))

                        if not tpc:
                            # fail over
                            t = tp.getName()
                        else:
                            t = tpc

			self.typecode.append('%s(pname=(ns,"%s"),aname="_%s",minOccurs=0)' \
                                             %(t, part.getName(),
                                               part.getName()))
		    elif tp.getName():

                        self.docCode.append(part.getName())
                        self.docCode.append(tp.getName())
			self.typecode.append('%s(pname=(ns,"%s"),aname="_%s",minOccurs=0)' \
                                             %(tp.getName(), part.getName(),
                                               part.getName()))
		    else:
			raise WsdlGeneratorError, 'shouldnt happen'

		elif tp.isComplexType():
                    self.docCode.append(part.getName())
                    self.docCode.append(tp.getName())
		    self.typecode.append('%s( name="%s", ns=ns )'\
                                         %(tp.getName(), part.getName()))

		else:
                    raise WsdlGeneratorError, 'shouldnt happen'
		return


class SchemaDescription:
    """Generates classes for all global definitions and declarations in 
       a schema instance.
    """
    def __init__(self):
        self.nsh = NamespaceHash()
        self.typeDict = {}
        return

    def fromWsdl(self, schema, alternateWriter):
        """schema -- schema.Schema instance
        """
        if not isinstance(schema, ZSISchemaAdapter):
	    raise TypeError, 'type %s not a Schema' %(schema.__class__)

	self.header = '%s \n# %s \n#\n# %s \n%s\n' \
                      %('#'*30, 'targetNamespace',
                        schema.getTargetNamespace(), '#'*30)

        self.header += '\n\n# imported as: %s' % \
                       self.nsh.getAlias(schema.getTargetNamespace())

        self.header += '\nclass %s:' % \
                       self.nsh.getModuleName(schema.getTargetNamespace())
        self.header += "\n%stargetNamespace = '%s'" % \
                       (ID1, schema.getTargetNamespace())

	self.body = ''
        self.last = ''
        
        class_dict = {}
        class_list = []
        
        self.generate(schema.getTypesDict(), alternateWriter, class_list, class_dict)
        self.generate(schema.getElementsDict(), alternateWriter, class_list, class_dict)
        self.getClassDefs(class_list, class_dict)
        
        self.body += '\n\n# define class alias for subsequent ns classes'
        self.body += '\n%s = %s' \
                     % ( self.nsh.getAlias(schema.getTargetNamespace()),
                         self.nsh.getModuleName(schema.getTargetNamespace()))
        self.body += '\n_add_namespace_container(%s)' %self.nsh.getModuleName(schema.getTargetNamespace())
        if self.last:
            self.body += self.last

    def generate(self, sdict, alternateWriter, class_list, class_dict):

        if alternateWriter:
            exec( 'import %s' % alternateWriter[0] )
            alternateWriter = '%s.%s()' % (alternateWriter[0],
                                           alternateWriter[1] )
        elements = [] 
        typedefs = [] 
        #Sort types/elements 
	for tp in sdict.values():
            #tw = self.TypeWriter()
            #ref = weakref.ref(self)
            #tw.fromType(tp, ref)

            #JRB
            if tp._type.isDefinition():
                typedefs.append(tp)            
            elif tp._type.isElement():
                elements.append(tp)            
            else:
                raise RuntimeError, 'Not Element or Definition?'

        #Sort types
        typeitems = [] 
        deriveditems = [] 
	for tp in typedefs:
            if tp._type.content is None:
                typeitems.append(tp)
            elif tp._type.content.isModelGroup():
                typeitems.append(tp)
            elif tp._type.content.isComplex() or tp._type.content.isSimple():
                    base = tp._type.content.derivation.getTypeDefinition('base')
                    if base is None:
                        typeitems.append(tp)
                    elif base.getTargetNamespace() != tp._type.getTargetNamespace():
                        typeitems.append(tp)
                    elif _is_xsd_or_soap_ns(base.getTargetNamespace()):
                        # JRB Could screw up if we are extending one of these namespaces
                        typeitems.append(tp)
                    else:
                        deriveditems.append(tp)
            elif tp._type.content.isRestriction() or tp._type.content.isExtension():
                deriveditems.append(tp)
            else:
                raise RuntimeError, 'Not supporting content %s' %tp._type.content

	for tp in typeitems + deriveditems:
            tw = self.TypeWriter()
            ref = weakref.ref(self)
            tw.fromType(tp, ref)
            #self.body += "=====> generate 1"
            #self.body += tw.extractCode()
            #self.body += "<====="
            class_list.append(tp)
            class_dict[tp] = [tw]

        elementitems = [] 
        deriveditems = [] 
        for tp in elements:
            if tp._type.getAttribute('ref') is None:
                elementitems.append(tp)
            else:
                deriveditems.append(tp)
                    

	for tp in elementitems + deriveditems:
            tw = self.TypeWriter()
            ref = weakref.ref(self)
            tw.fromType(tp, ref)
            class_list.append(tp)
            class_dict[tp] = [tw]

        for item in class_list:
            tw = class_dict[item][0]
            if tw.precede:
                if class_dict.has_key(tw.precede):
                    class_dict[tw.precede].append(tw)
                else:
                    pass
 
    def getClassDefs(self, class_list, class_dict):
        check_list = []
        for indx in range(len(class_list)):
            if class_dict.has_key(class_list[indx]):
                for tw in class_dict[class_list[indx]]:
                    self.body += tw.extractCode()
                    check_list.append(tw.name)
                else:
                    del class_dict[class_list[indx]]

        # JRB: I really dont know how much of this is
        # useful code, but I hacked in an ordering 
        # for types before declarations.
        if len(check_list) > 0:
            do_types = []
            do_elements = []
            for className in check_list:
                if className.endswith('_Def'):
                    do_types.append(className)
                else:
                    do_elements.append(className)
            self.getClassDefs(do_types, class_dict)
            self.getClassDefs(do_elements, class_dict)
        else:
            for l in class_dict.values():
                for tw in l:
                    #self.body += "=====> getClassDefs 2"
                    self.body += tw.extractCode()
                    #self.body += "<====="


    def extractCode(self, tw):
        self.body += tw.prepend.getvalue()
        self.body += tw.classdef.getvalue()
        self.body += tw.classvar.getvalue()
        self.body += tw.initdef.getvalue()
        self.body += tw.initcode.getvalue()
        self.body += tw.basector.getvalue()
        self.body += tw.postpend.getvalue()


    def write(self, fd=sys.stdout):
	fd.write(self.header)
	fd.write(self.body)
	fd.write('\n'*4)

    class TypeWriter:
        """Generates a string representation of a typecode representing
           a schema declaration or definition.
        """
	def __init__(self):
	    self.bti = BaseTypeInterpreter()
            self.nsh = NamespaceHash()
	    self.name = None
	    self.precede = None
            self.prepend  = StringWriter()
            self.classdef = StringWriter()
            self.classvar = StringWriter()
            self.initdef  = StringWriter()
            self.initcode = StringWriter()
            self.basector = StringWriter()
            self.postpend = StringWriter()
            self.allOptional = False
            self.hasRepeatable = False
            self.typeList  = []
            self.typeDict  = {}
            self.localDefs = []
	    return

        def extractCode(self):
            formattedType = ''
            formattedType += self.prepend.getvalue()
            formattedType += self.classdef.getvalue()
            formattedType += self.classvar.getvalue()
            formattedType += self.initdef.getvalue()
            formattedType += self.initcode.getvalue()
            formattedType += self.basector.getvalue()
            formattedType += self.postpend.getvalue()
 
            formattedType += self.extractSubtypes()
             
            return formattedType

        def extractSubtypes(self):
            subTypes = ''
 
            for t in self.localDefs:
                subTypes += t.extractCode()

            formatted = []

            for l in string.split(subTypes, '\n'):
                if l:
                    formatted.append('%s%s' % (ID1, l))
                else:
                    formatted.append(l)
                    
            return '\n'.join(formatted)

	def fromType(self, myType, parentRef):
            """myType -- Type representation
            """

	    tp = myType

            if tp.isSimpleType():
                self.name = tp.getName() + '_Def'
                self._fromSimpleType(tp)

	    elif tp.isWildCard():
                self._fromWildCard(tp)

            elif tp.isElement():
                self.name = tp.getName() + '_Dec'
                self._fromElement(tp)

	    elif tp.isComplexType():
                self.name = tp.getName() + '_Def'
                self._fromComplexType(tp)

            elif tp.isAttribute():

                self._fromAttribute(tp)

	    else:
		raise WsdlGeneratorError, 'WARNING: NOT HANDLED %s' \
                      % (tp.__class__)

            alias = self.nsh.getAlias(tp.getTargetNamespace())
            key = "%s.%s_Def" % (alias, tp.getName())
            if self.typeList:
                self.typeList.sort()
                    # add entry to type dictionary for later use in
                    # docstring generation
                self.typeDict[key] = self.typeList
	    return

        def _fromSimpleType(self, tp):
            
            tp = tp.getDefinition()
            
            if tp.getName():
                tpc = tp.getTypeclass()
                self.initdef.set('\n%sdef __init__(self, pname, **kw):' % (ID2))
                objName = '_' + tp.getName()
                if tpc:
                    typeName = self.bti.get_pythontype(None, None, tpc)
                    if not typeName:
                        typeName = 'Any'
                    self.precede  = '%s' % (tpc)
                    className = '%s_Def' %tp.getName()
                    self.classdef.set('\n\n%sclass %s(%s, TypeDefinition):' \
                                      % (ID1, className, tpc))
                    #self.initcode.set('\n%saname = None' % ID3 )
                    #self.initcode.write('\n%sif name:' % ID3)
                    #self.initcode.write('\n%skw["pname"] = name' \
                    #                    % ID4)
                    #self.initcode.write('\n%skw["aname"] = "_%%s" %% name' \
                    #                    % ID4)
                    self._handleAttributes(tp._def, className)
                    self.basector.set('\n%s%s.__init__(self, pname, **kw)'\
                                      % (ID3,tpc))
                else:
                    typeName = 'Any'
                    # XXX: currently, unions will get shuffled thru here.
                    self.classdef.set('\n\n%sclass %s(ZSI.TC.Any):' \
                                      % (ID1, tp.getName() + '_Def'))
                    self.initcode.write('\n%s# probably a union - dont trust it' % ID3)
                    self.basector.set('\n%sZSI.TC.Any.__init__(self, pname=name, aname="_%%s" %% name, minOccurs=0, maxOccurs="unbounded", **kw)' % ID3)

                self.typeDoc('minOccurs=0', objName, typeName)
            else:
                raise WsdlGeneratorError, 'shouldnt happen'

        def _fromWildCard(self, tp):
            # XXX: not particularly trustworthy either.  pending further work.
            tp = tp.getDeclaration()
            self.classdef.set('\n\n%sclass %s(ZSI.TC.XML):' \
                              % (ID1, tp.getName()))
            self.initdef.set('\n%s__init__(self,pname):' % (ID2))
            self.basector.set('\n%sZSI.TC.XML.__init__(self,pname,**kw)' % ID3)

        def _fromAttribute(self, tp):
            self.classdef.set('\n\n%sclass %s:' % (ID1, tp.getName()))
            self.classvar.set('\n%s# not yet implemented' % ID2)
            self.classvar.write('\n%s# attribute declaration' % ID2)
            self.classvar.write('\n%spass\n' % ID2)

        def _fromElement(self, tp):
            etp = tp.getType()

            if etp is None:
                self._writeElementComplexEmpty(tp)
            elif etp and etp.isDefinition():    
                if etp.isSimpleType():
                    self._elementSimpleType(tp, etp)
                elif etp.isComplexType():
                    self._elementComplexType(tp, etp)
            else:
                raise WsdlGeneratorError, 'Unknown type(%s) not handled ' \
                      % (etp.__class__)

        def _elementSimpleType(self, tp, etp):

            tpc = etp.getTypeclass()
            self.precede   = '%s' % (tpc)

            self.classdef.set('\n\n%sclass %s(%s, ElementDeclaration):' \
                              % (ID1, tp.getName() + '_Dec', tpc))
            
            self.classvar.set('\n%sliteral = "%s"' % (ID2, tp.getName()))
            self.classvar.write('\n%sschema = "%s"' % \
                                (ID2,tp.getTargetNamespace()))
            
            self.initdef.set('\n\n%sdef __init__(self, **kw):' \
                             % ID2)

            self.initcode.set('\n%spname = self.get_pname()'% ID3)
            self.initcode.write('\n%saname = self.get_aname(pname[1])'% ID3)

            self.basector.set('\n\n%s%s.__init__(self,pname=pname,aname=aname,**kw)' % (ID3,tpc))
            typeName = self.bti.get_pythontype(None, None, tpc)
            self.typeDoc('', '__param', typeName)
                  

        def _elementComplexType(self, tp, etp):
            '''tp -- element declaration
               etp -- type definition
            '''
            if etp.getName():

                self.precede  = '%s' %(etp.getName() + '_Def' )
                
                if etp.getTargetNamespace() != tp.getTargetNamespace():

                    nsp = etp.getTargetNamespace()
                    self.classdef.set('\n\n%sclass %s(%s.%s,ElementDeclaration):' \
                                      % (ID1, tp.getName() + '_Dec',
                                         self.nsh.getAlias(nsp),
                                         etp.getName() + '_Def'))
                else:
                    self.classdef.set('\n\n%sclass %s(%s,ElementDeclaration):' \
                                      % (ID1, tp.getName() + '_Dec',
                                         etp.getName() + '_Def'))

                self.classvar.set('\n%sliteral = "%s"' %(ID2, tp.getName()))
                self.classvar.write('\n%sschema = "%s"'%(ID2,tp.getTargetNamespace()))
                self.initdef.set('\n\n%sdef __init__(self, **kw):' %(ID2))
                self.initcode.set('\n%spname = self.get_pname()'% ID3 )

                nsp = etp.getTargetNamespace()
                
                typeName = '%s.%s_Def' % (self.nsh.getAlias(nsp), etp.getName())
                self.basector.set('\n\n%s%s.__init__(self, pname=pname, **kw)'  % (ID3, typeName))
                self.typeDoc('', '_' + tp.getName(), typeName)
            elif etp._def.content is None:
                self._writeElementComplexEmpty(tp)
            #elif tp.getName() == 'createCounterResponse':
            #    raise RuntimeError, 'WHAT IS GOING ON.'            
            else:
                # <element name='ged'><complexType>...</complexType></element>
                self._fromComplexType(etp.expressLocalAsGlobal(tp))

                # now we are discaring the _Dec(LOCAL_Def) subclassing
                # and expressing the declaration w/the local def as
                # a def.  retweek the classdef and off we go...
                #self.classdef = StringWriter()
                self.classdef.set('\n\n%sclass %s_Dec(ZSI.TCcompound.Struct, ElementDeclaration):' %(ID1, tp.getName()))
                self.initdef.set('\n%sdef __init__(self, **kw):' % (ID2))
                #JRB Big Hack (schema,"whatever") ==> "whatever"
                subclassvar = _getLiteralFromTypeClassVar(self.classvar.getvalue())
                #self.classvar.set(re.sub('type', 'literal', subclassvar))
                self.classvar.set(subclassvar)
                #self.initcode.set('\n%sname = name or self.__class__.literal\n%sns = ns or self.__class__.schema\n%s' % ( ID3, ID3, self.initcode.getvalue()))
                self.initcode.write('\n%spname = self.get_pname()\n' %ID3)

        def _writeElementComplexEmpty(self, tp):
            '''<element name='whatever'><complexType/></element>
            '''
            #JRB ?? <element name='whatever'/>
            self.classdef.set('\n\n%sclass %s_Dec(ZSI.TCcompound.Struct, ElementDeclaration):' \
                              % (ID1, tp.getName()))
            self.classvar.set('\n%sschema = "%s"' %(ID2,tp.getTargetNamespace()))
            self.classvar.write('\n%sliteral = "%s"'%(ID2,tp.getName()))
            self.initdef.set('\n%sdef __init__(self, **kw):' % (ID2))
            self.initcode.set('\n%spname = self.get_pname()'% ID3)
            self.initcode.write('\n%saname = self.get_aname(pname[1])'% ID3)
            #self.basector.set('\n%sZSI.TCcompound.Struct.__init__(self, self.__class__, [], pname=pname, aname=aname, inline=1)' %ID3)
            # MIXED
            mixed = ''
       
            if tp._type.content is not None and tp._type.content.getAttribute('mixed') == 'true':
                mixed=',mixed=True, mixed_aname="_text"'
            self.basector.set('\n%sZSI.TCcompound.Struct.__init__(self, self.__class__, [], pname=pname, aname=aname, inline=1%s)' %(ID3,mixed))

        def _fromComplexType(self, tp):

            if isinstance(tp, ZSI.wsdlInterface.ZSISchemaTypeAdapter ):
                # the "usual"
                tp = tp.getDefinition()
            else:
                # this is when an element has
                # a local complex type def
                #pass
                print_debug("_fromComplexType tp definition -- %s" %tp._def)
                
            typecodelist = '['


            if tp.isComplexContent():
                self._complexTypeComplexContent(tp)
                return

            if tp.isSimpleContent():
                self._complexTypeSimpleContent(tp)
                return 

            # ok, it's not derived content and therefore has a model group
            # write out the class def and class variables

            className =  '%s_Def' %tp.getName()
            self.classdef.set('\n\n%sclass %s:' %(ID1, className))
            
##             if self._complexTypeHandleAttributes(tp):
##                 # not yet implemented
##                 pass

            self.classvar.set("\n%sschema = '%s'" % (ID2,
                                                     tp.getTargetNamespace()))
            self.classvar.write("\n%stype = (schema,'%s')" % (ID2, tp.getName()))

            self._handleAttributes(tp._def, className)

            self.initdef.set('\n%sdef __init__(self, pname, **kw):' % (ID2))

            typecodelist = '['
            mg = tp.getModelGroup()
            if mg.isAll() or mg.isSequence():
                typecodelist += self._complexTypeAllOrSequence(tp, mg)
            elif mg.isChoice():
                typecodelist += self._complexTypeChoice(tp, mg)
            else:
                # if we get here, we hit a "void" model group.  ie:
                # <xsd:element name="getValue">
                #  <xsd:complexType/>
                # </xsd:element>
                self.classdef.set('\n\n%sclass %s(ZSI.TCcompound.Struct):'\
                                  %(ID1, tp.getName() + '_Def'))

            typecodelist += ']'
            self._complexTypecodeLogic( typecodelist, tp.getTargetNamespace(), tp)

        def _complexTypeComplexContent(self, tp):

            dt = tp.getDerivedTypes()
            tc = dt.getTypeclass()

            if not tc:
                # ie: not in a default namespace
                if dt.getDerivation():
                    # XXX: this will need more work

                    # ok, i'm totally cheating here.  getDerivedContent
                    # returns a ModelGroupAdapter even tho it's not
                    # technically a model group.  however, i can then
                    # feed it to the complex type generation methods to
                    # harvest a type code list.  to end the cheat,
                    # self.initdef|classdef are re-assigned to here and
                    # off we go.
                    if dt.contentIsSequence() or dt.contentIsAll():
                        tclist = self._complexTypeAllOrSequence(tp,
                                                                dt.getContent())
                    elif dt.contentIsChoice():
                        tclist = self._complexTypeChoice(tp,
                                                         dt.getContent())
                    else:
                        # dubious?
                        tclist = ''
                        
                    # end cheating....

                    self.precede  = '%s%s' % ( dt.getDerivation(), '_Def' )
                    nsp = self.nsh.getAlias(tp.getTargetNamespace())

                    # JRB Busted: Need to grab bases tagetNamespace and
                    # resolve to alias
                    base = dt._content.derivation.getTypeDefinition('base')
                    dnsp = nsp
                    className = '%s_Def' %tp.getName()
                    if base is not None:                    
                        dnsp = self.nsh.getAlias(base.getTargetNamespace())
                        if dnsp == nsp:
                            self.classdef.set('\n\n%sclass %s(%s_Def, TypeDefinition):' %(ID1, className, dt.getDerivation()))
                        else:
                            self.classdef.set('\n\n%sclass %s(%s.%s_Def, TypeDefinition):' %(ID1, className, dnsp, dt.getDerivation()))
                    else:
                        self.classdef.set('\n\n%sclass %s(%s_Def, TypeDefinition):' %(ID1, className, dt.getDerivation()))

                    self.classvar.set('\n%s# rudimentary - more soon' % ID2)
                    self.classvar.write("\n%sschema = '%s'" % \
                                        (ID2, tp.getTargetNamespace()))
                    self.classvar.write("\n%stype = (schema,'%s')\n" % (ID2,tp.getName()))
                    self._handleAttributes(tp._def, className)
                    self.initdef.set('\n\n%sdef __init__(self, pname, **kw):' % ID2)
                    #self.initcode.write('\n\n%sif name:' % ID3)
                    self.initcode.write('\n%sTCList = [%s]' % (ID3, tclist))

                    #JRB Busted
                    self.basector.set('\n%s%s.%s_Def.__init__(self, pname, **kw)' % (ID3, dnsp, dt.getDerivation()))
                    if dt.isExtension():
                        # JRB Only attributes the extension declares will be
                        # found.  What we really need to do is copy the 
                        # attribute_type_dict of the base and add any 
                        # attributes found in the extension.
                        self.postpend.set('\n%s# extending....' % ID3)
                        self.postpend.write('\n%sself.ofwhat += tuple(TCList)'\
                                            % ID3)
                        self._handleAttributes(dt._content.derivation, className)
                    elif dt.isRestriction():
                        self.postpend.set('\n%s# restricting....' % ID3)
                        self.postpend.write('\n%sself.ofwhat = tuple(TCList)' \
                                            % ID3)
                        #self.postpend.write('\n%sself.lenofwhat = len(TCList)' % ID4)
                else:
                    self.classdef.set('\n\n%sclass %s:' \
                                      % (ID1, tp.getName() + '_Dec'))
                    self.classvar.set('\n%s# not yet implemented' % ID2)
                    self.classvar.write('\n%s# non array complexContent' % ID2)
                    self.classvar.write('\n%spass\n' % ID2)
            elif '%s' % tc == 'ZSI.TCcompound.Array':
                # ladies and gents, we have an array
                self.classdef.set('\n\n%sclass %s(%s, TypeDefinition):' \
                                  % (ID1, tp.getName() + '_Def', tc))
                self.initdef.set("\n%sdef __init__(self, name = None, ns = None, **kw):" % ID2)

                arrayinfo = dt.getArrayType()
                
                if arrayinfo:
                    nsp = ''
                    if arrayinfo[2]:
                        nsp  = self.nsh.getAlias(tp.getTargetNamespace())
                        nsp += '.'
                        atype = arrayinfo[1] + '_Def'
                        typeName = '%s%s' % (nsp, atype)
                    else:
                        atype = arrayinfo[1]
                        typeName = self.bti.get_pythontype(None, None, atype)
                        if not typeName:
                            typeName = 'Any'

                    self.basector.set("\n%s%s.__init__(self, '%s', %s%s(name=None), pname=name, aname='_%%s' %% name, oname='%%s xmlns=\"%s\"' %% name, **kw)" \
                                      % (ID3, tc, arrayinfo[0], nsp,
                                         atype, tp.getTargetNamespace()))
                        
                    self.typeDoc('', '_element', typeName)
                else:
                    raise WsdlGeneratorError, 'failed to handle array!'
            elif '%s' % tc == 'ZSI.TC.Any':
                # this is a 'special case' - it's possible in schema to
                # extend an xsd:anyType - anyType is not currently
                # really well supported, this is something of a
                # preliminary fix.
                className = '%s_Def' %tp.getName()
                self.classdef.set('\n\n%sclass %s(ZSI.TCcompound.Struct, TypeDefinition):' %(ID1, className))
                self.classvar.set('\n%s# short term fix for derived anyType' \
                                  % ID2)
                self.classvar.write("\n%sschema = '%s'" % \
                                    (ID2, tp.getTargetNamespace()))
                self.classvar.write("\n%stype = (schema,'%s')" % (ID2,tp.getName()))
                self._handleAttributes(tp._def, className)
                self.initdef.set('\n\n%sdef __init__(self, pname, **kw):' % ID2)
                typecodelist = '[ZSI.TC.Any(*kw), ]'
                self._complexTypecodeLogic(typecodelist, tp.getTargetNamespace(), tp)
                
            else:
                raise WsdlGeneratorError, 'failed to handle complex content'
            
            return

        def _complexTypeSimpleContent(self, tp):
            dt = tp.getDerivedTypes()

            if dt.isSimpleContent() and dt.getTypeclass():
                # XXX: this is also catching extension/restriction
                # of simple types and will need some attention later
                className = '%s_Def' %tp.getName()
                self.classdef.set('\n\n%sclass %s(%s, TypeDefinition):' % (ID1, className, dt.getTypeclass()))
                self.classvar.set('\n%s# rudimentary support' % ID2)
                self.classvar.write('\n%sschema = "%s"' %(ID2,tp.getTargetNamespace()))
                self.classvar.write('\n%stype = (schema,"%s")' %(ID2,tp.getName()))
                self._handleAttributes(tp._def, className)
                self.initdef.set('\n%sdef __init__(self, pname, **kw):' % ID2)
                self.basector.set('\n%s%s.__init__(self, pname=pname, **kw)' \
                                  %( ID3, dt.getTypeclass()))
                return
            elif dt.isRestriction() or dt.isExtension():
                self.precede  = '%s%s' % ( dt.getDerivation(), '_Def' )
                nsp = self.nsh.getAlias(tp.getTargetNamespace())
                className = '%s_Def' %tp.getName()
                self.classdef.set('\n\n%sclass %s(%s, TypeDefinition):' %(ID1, className, dt.getDerivation() + '_Def'))
                self.classvar.set('\n%s# rudimentary support' % ID2)
                #self.classvar.write('\n%stag = "%s"' % (ID2, tp.getName()))
                self.classvar.write('\n%sliteral = "%s"' %(ID2, tp.getName()))
                self.classvar.write('\n%sschema = "%s"' % \
                                    (ID2, tp.getTargetNamespace()))
                self._handleAttributes(tp._def, className)
                
                self.initdef.set('\n\n%sdef __init__(self, name=None, ns=None, **kw):' % ID2)
                self.initcode.set('\n%sname = name or self.__class__.literal'\
                                  % ID3)
                self.initcode.write('\n%sns = ns or self.__class__.schema' \
                                    % ID3)
                self.basector.set('\n\n%s%s.__init__(self, name=name, ns=ns, **kw)' \
                                  %( ID3, dt.getDerivation() + '_Def'))
            else:
                raise WsdlGeneratorError, \
                      'could not determine simple content base'


        def _complexTypeAllOrSequence(self, tp, mg):
            if tp._def.isDefinition() and not isinstance(tp._def, LocalComplexType):
                classDefName = '%s_Def' %tp.getName()
            else:
                classDefName = '%s_Dec' %tp.getName()

            self.classdef.set('\n\n%sclass %s(ZSI.TCcompound.Struct, TypeDefinition):'\
                              %(ID1, classDefName))

            typecodelist = ''

            self.initcode.set('\n%s# complexType all or sequence' % ID3)

            for e in mg.getContent():
                if e.isDeclaration() and e.isElement():

                    etp = None
                            
                    if e.getType():
                        etp = e.getType()
         
                    if e.getName():
                        self.initcode.write('\n%sself._%s = None' \
                                            % (ID3, e.getName()))
                    elif etp and etp.getName():
                        # element references
                        self.initcode.write('\n%sself._%s = None' \
                                            % (ID3, etp.getName()))

                    if e.getName():
                        objName = '_' + e.getName()
                    else:
                        objName = '_any'   # fix

                    #occurs = None
                    minOccurs,maxOccurs,nillable = self._getOccurs(e)
                    if e.isDeclaration() and e.isWildCard():
                        #occurs = self._calculateOccurance(e)

                        typeName = 'AnyElementXXX WHO KNOWS WHAT THIS IS FOR?'
                        #typecodelist += self._getWildcardTypecode(e, occurs)
                        typecodelist += self._getWildcardTypecode(e)

                    elif e.isDeclaration() and e.isAnyType():
                        typeName = 'AnyType'
                        typecodelist  += 'ZSI.TC.AnyType(pname="%s",aname="_%s",minOccurs=%s,maxOccurs=%s,nillable=%s), '\
                                         %(e.getName(),e.getName(),minOccurs,maxOccurs,nillable)

                    elif e.isDeclaration() and e.isElementReference():
                        #occurs = self._calculateOccurance(e)
                        nsp = self.nsh.getAlias(etp.getTargetNamespace())
                        typeName = '%s.%s_Dec' % (nsp, e.getName())
                        #JRB
                        typecodelist +='%s(aname="_%s",minOccurs=%s,maxOccurs=%s,nillable=%s),'\
                                        %(typeName,e.getName(),minOccurs,maxOccurs,nillable)

                    elif e.isDeclaration() and e.containsLocalDefinition():
                        typeName = ''
                        localType = self.__class__()
                        localType._fromComplexType(e.getLocalDefinition())
                        self.localDefs.append(localType)
                        typecodelist += 'self.__class__.%s(name="%s",ns=ns,minOccurs=%s,maxOccurs=%s,nillable=%s), ' \
                            %(e.getName() + '_Def', e.getName(), minOccurs,maxOccurs,nillable)

                    elif etp.isDefinition() and etp.isSimpleType():
                        tpc = None
                        tpc = etp.getTypeclass()

                        if tpc and e.isXSD():
                            typeName = self.bti.get_pythontype(None, None, tpc)
                            if not typeName:
                                typeName = 'Any'
                            typecodelist +=\
                                         '%s(pname=(ns,"%s"),aname="_%s",minOccurs=%s,maxOccurs=%s,nillable=%s), ' \
                                         %(tpc,e.getName(),e.getName(),minOccurs,maxOccurs,nillable)
                        else:
                            nsp = self.nsh.getAlias(etp.getTargetNamespace())
                            typeName = '%s.%s' % (nsp, etp.getName()) + '_Def'
                            typecodelist +='%s(pname=(ns,"%s"),aname="_%s"), '\
                                            %(typeName, e.getName(), e.getName())

                    elif etp.isDefinition() and etp.isComplexType():
                        nsp = self.nsh.getAlias(etp.getTargetNamespace())
                        typeName = '%s.%s' % (nsp,etp.getName()) + '_Def'
                        typecodelist  += '%s(pname=(ns,"%s"), aname="_%s",minOccurs=%s,maxOccurs=%s,nillable=%s), ' \
                                         %(typeName,e.getName(),e.getName(),minOccurs,maxOccurs,nillable)
                        self.precede = '%s_Def' %etp.getName()

                    elif etp:
                        # here we have a hold of a LocalElementDef
                        # like an element reference.  bogosity.
                        # we dont hit this much
                        typeName = etp.getType().getName()
                        typecodelist  += '%s(pname=(ns,"%s"),aname="_%s",minOccurs=%s,maxOccurs=%s,nillable=%s), '\
                                         % (typeName, etp.getName(), etp.getName(),minOccurs,maxOccurs,nillable)
                    elif e:
                        raise WsdlGeneratorError, 'instance %s not handled '\
                              % (e.getName())
                    else:
                        raise WsdlGeneratorError, 'instance %s not handled '\
                              % (e)
                    
                    if minOccurs is not None:
                        self.typeDoc(minOccurs + '/' + maxOccurs + '/' + nillable, objName, typeName)
                    else:
                        self.typeDoc(None, objName, typeName)

                elif e.isDeclaration() and e.isElementReference():
                    #occurs = self._calculateOccurance(e)
                    nsp = self.nsh.getAlias(etp.getTargetNamespace())
                    typeName = '%s.%s' % (nsp, e.getName()) + '_Dec'
                    typecodelist +='%s(ns=ns,minOccurs=%s,maxOccurs=%s,nillable=%s), '\
                                    %(typeName,minOccurs,maxOccurs,nillable)
                else:
                    raise WsdlGeneratorError, 'instance %s not handled '\
                          % (e.__class__)
            else:
                pass
            #
            nsp = self.nsh.getAlias(tp.getTargetNamespace())
            self.initcode.write('\n%sns = %s.%s.get_nsuri()'%(ID3, nsp, classDefName))

            return typecodelist

        def _complexTypeChoice( self, tp, mg ):

            # XXX: this has been somewhat neglected

            self.allOptional = True

            return self._complexTypeAllOrSequence( tp, mg )

            """ JRB This appears to be totally not used? XXX

            # the rest of this is pending further work, using
            # the All/Sequence gives us immediate functionality.

            # we shall get back to this....

            typecodelist = 'ZSI.TCcompound.Choice(['

            self.initcode.set('\n%s# internal vars' % ID3)
		    
            for e in mg.getContent():
                        
                if e.isDeclaration() and e.isElement():
                            
                    etp = None
                    if e.getType():
                        etp = e.getType()
				
                    self.initcode.write('\n%sself._%s = None' \
                                        % (ID2, e.getName()))
                            
                    if e.isAnyType():
                        typecodelist += 'ZSI.TC.Any(pname="%s",aname="%s"), '\
                                        %(e.getName(),e.getName())

                    elif etp.isDefinition() and etp.isSimpleType():
                        occurs = self._calculateOccurance(e)

                        tpc = None
                        tpc = etp.getTypeclass()

                        typecodelist  += '%s(pname="%s",aname="_%s"%s), '\
                                         %(tpc,e.getName(),e.getName(),
                                           occurs)

                    elif etp.isDefinition() and etp.isComplexType():
                        typecodelist  += 'Struct(%s,%s().typecode,pname="%s",aname="_%s"), '\
                                         %(etp.getName(),etp.getName(),
                                           e.getName(),e.getName())
			    
                    elif etp:
                        self.initcode.write('\n%sself._%s = None' % \
                                            (ID2, e.getName()))
                        typecodelist  += '%s("%s"), '\
                                         %(etp.getName(), e.getName())
                    elif e:
                        raise WsdlGeneratorError, 'instance %s not handled '\
                              %(e.getName())
                    else:
                        raise WsdlGeneratorError, \
                              'instance %s not handled ' % (e)
                else:
                    raise WsdlGeneratorError, 'instance %s not handled '\
                          %(e.__class__)
            else:
                    pass

            return typecodelist
            """


        def _complexTypecodeLogic(self, typecodelist, tns, tp):
            # extra "smarts" for all regular (model group) complex type defs.

            # MIXED
            print_debug('_complexTypecodeLogic type definition\n%s' %tp._def)
            extraFlags = ''

            #ComplexType Definition
            if tp._def.getAttribute('mixed') == 'true':
                extraFlags = 'mixed=True,mixed_aname="_text",'
            elif tp._def.content is not None and tp._def.content.getAttribute('mixed') == 'true':
                extraFlags = 'mixed=True,mixed_aname="_text",'
            #if self.hasRepeatable:
            #    extraFlags += 'hasextras=1, '
            self.initcode.write('\n\n%sTClist = %s' % (ID3, typecodelist))
            self.basector.write('\n\n%sZSI.TCcompound.Struct.__init__(' % ID3)
            self.basector.write('self, self.__class__, TClist,')
            self.basector.write('\n%s%spname=pname, inorder=0,' % (ID3, ' ' * 31))
            self.basector.write('\n%s%s%s**kw)' % (ID3, ' ' * 31,
                                                   extraFlags))



        def _complexTypeHandleAttributes(self, tp):
            # XXX: need to revisit attributes - incomplete
            
            tmp = ''

            for a in tp.getAttributes():
                #tmp += '("%s","%s"),' % (a.getName(), a.getQName())
                pass

            return tmp
        """
        def _calculateOccurance(self, e):
            occurs = ''
            
            if e.getMaxOccurs() == 'unbounded':
                occurs += ', maxOccurs="unbounded"'
                self.hasRepeatable = True
            elif int(e.getMaxOccurs()) > 1:
                occurs += ', maxOccurs=%d' %int(e.getMaxOccurs)

            if int(e.getMinOccurs()) == 0  or self.allOptional:
                occurs += ', minOccurs=0'
            if e.isNillable():
                occurs += ', nillable=True'

            return occurs
        """

	def __recurse_tdc(self, tp):
	    tp.type
	    if isinstance(tp.type, TypeDescriptionComponent):
		tp.type
		tp = self.__recurse_tdc(tp.type)
	    else:
		return tp.type

        
        def typeDoc(self, occurs, objName, typeName):
            """Generates an entry in the type list for later use
               by docstring generation.
            """
            if not occurs:
                self.typeList.append([objName, typeName, False])
            elif occurs.find('minOccurs=0') == -1:
                self.typeList.append([objName, typeName, False])
            else:
                self.typeList.append([objName, typeName, True])

        def _getOccurs(self, e):
            '''e -- XMLSchema LocalElementDeclaration, ElementReference
            Currently this handles reference incorrectly, because the
            reference is not passed in (the element declaration is).
            '''
            nillable = e._dec.getAttribute('nillable')
            if nillable == 'true': nillable = '1'
            else: nillable = '0'
            maxOccurs = e._dec.getAttribute('maxOccurs')
            if maxOccurs == 'unbounded': maxOccurs = '"%s"' %maxOccurs
            minOccurs = e._dec.getAttribute('minOccurs')
            if self.allOptional is True:
                #JRB Hack
                minOccurs = '0'
                maxOccurs = '"unbounded"'
            return minOccurs,maxOccurs,nillable

        def _getWildcardTypecode(self, e):
            # this is totally dubious
            minOccurs,maxOccurs,nillable = self._getOccurs(e)
            if e.getName():
                return 'ZSI.TC.AnyType(pname=(ns,"%s"),aname="_%s",minOccurs=%d,maxOccurs=%s,nillable=%s), '\
                       %(e.getName(),e.getName(),minOccurs,maxOccurs,nillable)
            else:
                return 'ZSI.TC.AnyElement(aname="_any",minOccurs=%s,maxOccurs=%s,nillable=%s), '\
                       %(minOccurs,maxOccurs,nillable)

        def _handleAttributes(self, type, className=None):
            '''Only Definitions have 'mixed' content, dont want to
            specify mixed=False on declarations.
            '''
            if type.getAttribute('mixed') == 'true':
                self.classvar.write('\n%smixed = True' % ID2)

            if hasattr(type, 'getAttributeContent'):
                if type.getAttributeContent():
                    self.classvar.write('\n%sattribute_typecode_dict = {}' %ID2 )
                    if type.isExtension():
                        namespaceURI,extension = type.getAttribute('base')
                        extensionType = type.getTypeDefinition('base')
                        alias = None
                        if namespaceURI != type.getTargetNamespace():
                            alias = self.nsh.getAlias(namespaceURI)
                        if alias and extensionType.getAttributeContent():
                            self.postpend.write('\n%s%s.attribute_typecode_dict.update(%s.%s_Def.attribute_typecode_dict)' %(ID1,className,alias,extension))
                        elif extensionType.getAttributeContent():
                            self.postpend.write('\n%s%s.attribute_typecode_dict.update(%s_Def.attribute_typecode_dict)' %(ID1,className,extension))
                    attrs = ''
                    for a in type.getAttributeContent():
                        #if isinstance(a, ZSI.wstools.XMLSchema.AttributeWildCard):
                        if True is a.isWildCard() is a.isAttribute() is a.isDeclaration():
                            # XXX: pending further work
                            self.postpend.write('\n%s%s.attribute_typecode_dict[("%s","anyAttribute")] = ZSI.TC.AnyElement()' %(ID1,className,SCHEMA.XSD3))
                        else:
                            # Defined in Schema Instance
                            #t = a.getTypeDefinition('type')
                            # Built-in simpleType
                            t = a.getAttribute('type')
                            tc = self.bti.get_typeclass(t[1], t[0])
                            key = None
                            if a.getAttribute('form') == 'qualified':
                                key = '("%s","%s")' %(a.getTargetNamespace(),a.getAttribute('name'))
                            elif a.getAttribute('form') == 'unqualified':
                                key = '"%s"' %a.getAttribute('name')
                            else:
                                raise WsdlGeneratorError,\
                                    'attribute form(%s) must be unqualified or qualified'\
                                    %a.getAttribute('form')
                            #attrs += '%s:%s(), ' % (key,tc)
                            self.postpend.write('\n%s%s.attribute_typecode_dict[%s] = %s()' %(ID1,className,key,tc))
                    #self.postpend.write(' %s}' % attrs)
            self.postpend.write('\n')
